import numpy as np
from matplotlib import pyplot as plt
from matplotlib.collections import PatchCollection
import matplotlib.patches as patches
from matplotlib.colors import LogNorm
from ..misc.add_colorbar import add_colorbar
from .niceplot import niceplot
from ..SaxsAna.pysaxs3 import get_soq
from mpl_toolkits.axes_grid1 import make_axes_locatable
import copy


def shadeqrois(ax, qv, dqv, alpha=0.6, cmap='inferno', coords='data'):
    boxes = []
    
    # Loop over data points; create box from errors at each point
    cmap = plt.get_cmap(cmap)
    clrs = cmap(np.linspace(.1,.9,qv.size))
    if coords == 'axes':
        yl = ax.get_ylim()
    elif coords == 'data':
        xp, yp = ax.lines[-1].get_data()
        yl = (np.inf,0)
        for qi, dqi in zip(qv, dqv):
            q1 = qi - dqi/2
            q2 = qi + dqi/2
            y = yp[np.argmin(np.abs(q1 - xp)):np.argmin(np.abs(q2 - xp))]
            yl = (min([yl[0], y.min()]), max([yl[1], y.max()]))
    for qi, ci, dqi in zip(qv,clrs,dqv):
        q1 = qi - dqi/2
        q2 = qi + dqi/2
        rect = patches.Rectangle((q1, 2*yl[0]-yl[1]), dqi, (yl[1]-yl[0])*2)
        boxes.append(rect)

    # Create patch collection with specified colour/alpha
    pc = PatchCollection(boxes, facecolors=clrs, alpha=alpha,
                         edgecolor='k')

    # Add collection to axes
    ax.add_collection(pc)

def shade_wedges(ax, setup, alpha=0.6, cmap='inferno', qsec=(0,0), mirror=False):
    yl = ax.get_ylim()
    wedges = []
    r = setup.radii
    phiv = setup.phiv
    nr = len(r)
    nph = len(phiv)

    center = np.array(setup.center) - np.array(qsec[::-1])

    # Loop over data points; create box from errors at each point
    cmap = plt.get_cmap(cmap)
    clrs = cmap(np.linspace(0,1,nr*nph))

    for ri in r:
        for phi in phiv:
            w = patches.Wedge(center, ri[0], -(phi[0]+phi[1]), -phi[0] , width=ri[1])
            wedges.append(w)
            if mirror:
                w = patches.Wedge(center, ri[0], -(phi[0]+phi[1])-180, -phi[0]-180 , width=ri[1])
                wedges.append(w)

    # Create patch collection with specified colour/alpha
    pc = PatchCollection(wedges, facecolors=clrs, alpha=alpha,
                         edgecolor='k')

    # Add collection to axes
    lims = (ax.get_xlim(), ax.get_ylim())
    ax.add_collection(pc)
    ax.set_xlim(lims[0])
    ax.set_ylim(lims[1])

def plotqrois(Isaxs, setup, method='S(Q)', d=0, shade=False, color='r', ax=None,
              label='', mirror=False, cmapdata='inferno', cmaprois='viridis'):

    dim = Isaxs.shape

    if 'qsec' in vars(setup) and d==0:
        y1, x1 = setup.qsec[0]
        y2, x2 = setup.qsec[1]
    else:
        x1,x2 = ( max( setup.center[0]-d, 0 ), min( setup.center[0]+d, dim[1] ) )
        y1,y2 = ( max( setup.center[1]-d, 0 ), min( setup.center[1]+d, dim[0] ) )

    if ax is None:
        fig, ax = plt.subplots(1,1, figsize=(6,5), constrained_layout=True)

    if method == 'S(Q)' or method == 1:
        q, I, e  = get_soq(Isaxs, setup)
        ax.loglog(q, I, '.-', color=color, label=label)
        ax.set_xlabel(r'q [$\mathrm{nm}^{-1}$]')
        ax.set_ylabel(r'S(Q)')
        # niceplot(ax)
        if shade:
            shadeqrois(ax, setup.qv, setup.dqv)
        plt.tight_layout()
    elif method == 'ROIS' or method == 2:

        ## uncomment to check the masked qrois
        # Isaxs = Isaxs.copy()
        # for q in setup['qroi']:
        #     Isaxs[q[0],q[1]] = 1000

        uq = np.unique(setup.qv)
        cmaprois = plt.get_cmap(cmaprois, uq.size)
        flt = np.zeros_like(Isaxs)*np.nan

        saxs_sec = (Isaxs*setup.mask)[y1:y2,x1:x2]
        im = ax.imshow(saxs_sec, cmap=cmapdata, norm=LogNorm())
        for i,iuq in enumerate(uq):
            for j in np.where(setup.qv==iuq)[0]:
                qi = setup.qroi[j]
                flt[qi] = i
        im2 = ax.imshow(flt[y1:y2,x1:x2], cmap=cmaprois, alpha=.8)

        # shade_wedges(ax, setup, alpha=0.6, cmap='inferno', qsec=(y1,x1), mirror=mirror)

        # create an axes on the right side of ax. The width of cax will be 5%
        # of ax and the padding between cax and ax will be fixed at 0.05 inch.
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)

        cl = plt.colorbar(im, cax=cax)
        cl.ax.set_ylabel(r"intensity")
        ax.xaxis.set_visible(0)
        ax.yaxis.set_visible(0)

        add_colorbar(ax, setup.qv, label=r"$q\,(\mathrm{nm}^{-1})$", cmap=cmaprois,
                     discrete=True, location='top', tick_step=2, shrink=.8)

        # niceplot(ax, autoscale=False)
    plt.show()




import numpy as np
import matplotlib as mpl
from matplotlib import pyplot as plt

def add_colorbar(ax, vec, label=None, cmap='magma', discrete=False, tick_step=1,
                    qscale=0, location='right', show_offset=False, **kwargs):

    ncolors = len(vec)
    tick_indices = np.arange(0, len(vec), tick_step)
    vec = np.array(vec)[tick_indices]
    vec *= 10**qscale
    # divider = make_axes_locatable(ax)
    # cax = divider.append_axes('right', size='5%', pad=0.05)

    change_ticks = False
    if discrete:
        norm = mpl.colors.NoNorm()
        change_ticks = True
        if isinstance(cmap, str):
            cmap = plt.get_cmap(cmap, ncolors)
        elif isinstance(cmap, (list, np.ndarray)):
            cmap = mpl.colors.ListedColormap(cmap)
    else:
        cmap = plt.get_cmap(cmap)
        norm = mpl.colors.Normalize(vmin=vec.min(), vmax=vec.max())


    if location == 'right':
        orientation = 'vertical'
    elif location == 'top':
        orientation = 'horizontal'

    cax = mpl.colorbar.make_axes(ax, location=location, **kwargs)[0]
    cb = mpl.colorbar.ColorbarBase(cax, norm=norm, cmap=cmap,
                                    orientation=orientation)

    # set up color bar ticks
    if location == 'top':
        cax.xaxis.set_ticks_position('top')
        cax.xaxis.set_label_position('top')

    cb.set_label(label)
    if change_ticks:
        cb.set_ticks(tick_indices)
        cb.set_ticklabels(list(map(lambda x: "$%.{}f$".format(3-qscale) % x, vec)))
        if qscale and show_offset:
            cb.ax.text(1.,1.04, r'$\times 10^{{-{}}}$'.format(qscale),
                        transform=cb.ax.transAxes)
    cb.ax.invert_yaxis()
    cb.ax.set_in_layout(True)
    cax.grid(False)
    cb.ax.grid(False)




import numpy as np

def azimuthalAverage(image, center=None, stddev=False, returnradii=False, return_nr=False, 
        binsize=0.5, weights=None, steps=False, interpnan=False, left=None, right=None,
        mask=None ):
    """
    Calculate the azimuthally averaged radial profile.
    image - The 2D image
    center - The [x,y] pixel coordinates used as the center. The default is 
             None, which then uses the center of the image (including 
             fractional pixels).
    stddev - if specified, return the azimuthal standard deviation instead of the average
    returnradii - if specified, return (radii_array,radial_profile)
    return_nr   - if specified, return number of pixels per radius *and* radius
    binsize - size of the averaging bin.  Can lead to strange results if
        non-binsize factors are used to specify the center and the binsize is
        too large
    weights - can do a weighted average instead of a simple average if this keyword parameter
        is set.  weights.shape must = image.shape.  weighted stddev is undefined, so don't
        set weights and stddev.
    steps - if specified, will return a double-length bin array and radial
        profile so you can plot a step-form radial profile (which more accurately
        represents what's going on)
    interpnan - Interpolate over NAN values, i.e. bins where there is no data?
        left,right - passed to interpnan; they set the extrapolated values
    mask - can supply a mask (boolean array same size as image with True for OK and False for not)
        to average over only select data.
    If a bin contains NO DATA, it will have a NAN value because of the
    divide-by-sum-of-weights component.  I think this is a useful way to denote
    lack of data, but users let me know if an alternative is prefered...
    
    """
    # Calculate the indices from the image
    y, x = np.indices(image.shape)

    if center is None:
        center = np.array([(x.max()-x.min())/2.0, (y.max()-y.min())/2.0])

    r = np.hypot(x - center[0], y - center[1])

    if weights is None:
        weights = np.ones(image.shape)
    elif stddev:
        raise ValueError("Weighted standard deviation is not defined.")

    if mask is None:
        mask = np.ones(image.shape,dtype='bool')
    # obsolete elif len(mask.shape) > 1:
    # obsolete     mask = mask.ravel()

    # the 'bins' as initially defined are lower/upper bounds for each bin
    # so that values will be in [lower,upper)  
    nbins = int(np.round(r.max() / binsize)+1)
    maxbin = nbins * binsize
    bins = np.linspace(0,maxbin,nbins+1)
    # but we're probably more interested in the bin centers than their left or right sides...
    bin_centers = (bins[1:]+bins[:-1])/2.0

    # how many per bin (i.e., histogram)?
    # there are never any in bin 0, because the lowest index returned by digitize is 1
    #nr = np.bincount(whichbin)[1:]
    nr = np.histogram(r, bins, weights=mask.astype('int'))[0]

    # recall that bins are from 1 to nbins (which is expressed in array terms by arange(nbins)+1 or xrange(1,nbins+1) )
    # radial_prof.shape = bin_centers.shape
    if stddev:
        # Find out which radial bin each point in the map belongs to
        whichbin = np.digitize(r.flat,bins)
        # This method is still very slow; is there a trick to do this with histograms? 
        radial_prof = np.array([image.flat[mask.flat*(whichbin==b)].std() for b in xrange(1,nbins+1)])
    else: 
        radial_prof = np.histogram(r, bins, weights=(image*weights*mask))[0] / np.histogram(r, bins, weights=(mask*weights))[0]

    if interpnan:
        radial_prof = np.interp(bin_centers,bin_centers[radial_prof==radial_prof],radial_prof[radial_prof==radial_prof],left=left,right=right)

    if steps:
        xarr = np.array(zip(bins[:-1],bins[1:])).ravel() 
        yarr = np.array(zip(radial_prof,radial_prof)).ravel() 
        return xarr,yarr
    elif returnradii: 
        return bin_centers,radial_prof
    elif return_nr:
        return nr,bin_centers,radial_prof
    else:
        return radial_prof


import numpy as np

def fitline(x,y,dy):
    A = np.vstack((np.ones_like(x), x)).T
    C = np.diag(dy * dy)
    cov = np.linalg.inv(np.dot(A.T, np.linalg.solve(C, A)))
    b_ls, m_ls = np.dot(cov, np.dot(A.T, np.linalg.solve(C, y)))
    return (m_ls,np.sqrt(cov[1,1])),(b_ls,np.sqrt(cov[0,0]))

def fitline0(x,y,dy):
    m, dm = np.ma.average(y/x,weights=(dy/x)**(-2),returned=1)
    return m, 1/np.sqrt(dm)

def fitconstant(y, dy):
    b, db = np.ma.average(y,weights=dy**(-2),returned=1)
    return b, 1/np.sqrt(db)    



import os
import re
import pandas as pd
import numpy as np
from .Xfmt import Xfmt
from .to_h5 import to_h5
from ..misc.makemask import masker
import warnings

class Xdata(Xfmt):
    '''
    Class to get meta information on datasets based on data directory, headers and
    paths defined in Xfmt kernel.
    '''

    def __init__(self, **kwargs):
        super().__init__(kwargs.pop('fmtstr', ''))
        self.datdir = kwargs.pop('datdir', '')
        self._files = None
        self._masters = []
        self._headers = []
        self.meta  = []
        self._meta_save = None
        self._series = []
        self._series_ids = None

    def connect(self, datdir, **kwargs):
        """
        Finds data sets in `datdir` and reads meta data. Can be executed several times to append new data directories.
        :param datdir: data directory that contains data files
        :param kwargs:
        :return: None
        """
        if not os.path.isdir(datdir):
            warnings.warn('Data directory does not exist. Use valid data directory.')
            return
        self.datdir = os.path.abspath(datdir) + '/'
        if 'agipd' not in self.fmtstr:
            self._get_files(self.datdir)
            self._get_masters()
            self._get_headers()
            self._files2series()
        self._get_meta(**kwargs)

    def _get_files(self, datdir,):
        check_suffix = re.compile(self.suffix)
        files = [os.path.abspath(datdir + item) for item in os.listdir(datdir)
                    if os.path.isfile(os.path.join(datdir, item))
                    and bool(check_suffix.search(item))]
        self._files = sorted(files,
                          key=lambda x: int(''.join(re.findall(self.numfmt, x.split('/')[-1]))))

    def _get_masters(self):
        master = re.compile(self.masterfmt + r"\."+ self.suffix)
        masters = [master.search(x).group().split('/')[-1] for x in self._files
                   if bool(master.search(x))]
        self._masters = masters

    def _get_headers(self,):
        headers = []
        for m in self._masters:
            headers.append(self.get_header(self.datdir+m))
        self._header = headers

    def _files2series(self,):
        series = []
        series_id = []
        find_seriesid = re.compile(self.seriesfmt)
        for i, m in enumerate(self._masters):
            idstr = find_seriesid.search(m).group()
            series_id.append(int(idstr))
            nblocks = len(re.findall('(_\d{4,})', m))
            searchstr = idstr + '.*(_\d{{4,}}){{{}}}'.format(nblocks - 1)
            series.append([x for x in self._files if re.search(searchstr, x) is not None])
        self._series_ids = np.asarray(series_id, dtype='int32')
        self._series.extend(series)

    def _get_meta(self, addfirstnlast=True, checksubseries=True, nframesfromfiles=False):
        meta = {'series':self._series_ids, 'master':self._masters,
                'datdir':[self.datdir]*len(self._masters)}
        self.get_attributes(self, meta,)
        meta = pd.DataFrame.from_dict(meta)
        meta = meta.reindex(columns=['series']
                            + list([a for a in meta.columns
                                  if a not in ['series', 'master', 'datdir']])
                            + ['master', 'datdir'])
        if nframesfromfiles:
            for idx, row in meta.iterrows():
                row['nframes'] = len(self._series[idx])
                meta.loc[idx] = row

        if addfirstnlast:
            meta.insert(5, 'last', int(0))
            meta.insert(5, 'first', int(0))

            for idx, row in meta.iterrows():
                row['first','last'] = (0, int(row['nframes']-1))
                meta.loc[idx] = row
                if checksubseries:
                    tot_img = self.get_series(idx, verbose=False, output='shape')[0]
                    img_per_series = row['nframes']
                    nrow = row.copy()
                    idx_subset = 1
                    while nrow['last'] + 1 - tot_img < 0:
                        if 'subset' not in meta:
                            meta.insert(1, 'subset', int(0))
                        nrow['first'] = img_per_series + nrow['first']
                        nrow['last'] = img_per_series + nrow['last']
                        nrow['subset'] = idx_subset
                        idx_subset += 1
                        meta.loc[meta.shape[0]] = nrow


        if not len(self.meta):
            self.meta = meta
        else:
            self.meta = self.meta.append(meta, ignore_index=True)
            self.meta.drop_duplicates(inplace=True)
            self.meta.reset_index(drop=True, inplace=True)

    def get_series(self, series_id, **kwargs):
        """
        Reads data series.
        :param series_id: ID of series to analyze in Xdata.meta dataframe.
        :param kwargs: are passed to load_data_func in ProcData module.
        :return: np.ndarray if method is `full` or tuple of average images and variance if method is `average`.
        """
        if 'subset' in self.meta:
            nf = self.meta.loc[series_id, 'nframes']
            first = self.meta.loc[series_id, 'first']
            last = self.meta.loc[series_id, 'last']
            kwargs['first'] =  kwargs.get('first', first) % nf + first
            kwargs['last'] = kwargs.get('last', last) % nf + first
            series_id = self.meta[(self.meta['series']==self.meta.loc[series_id, 'series'])
                                    & (self.meta['subset']==0)].index.values[0]
        return self.load_data_func(self._series[series_id], xdata=self, **kwargs)

    def get_image(self, series_id, imgn=0, **kwargs):
        """
        Returns single image of dataset.
        :param imgn: Index of image starting with 0
        :param series_id: ID of series to analyze in Xdata.meta dataframe.
        """
        return self.load_data_func(self._series[series_id], xdata=self, first=(imgn,),
                                   last=(imgn+1,), **kwargs)[0]

    def masker(self, Isaxs, **kwargs):
        mask = kwargs.get('mask', self.setup.mask)
        masker(Isaxs, mask)

    def to_h5(self, series_id, filename, **kwargs):
        """
        Convert series to h5. Built in to convert single edf series from ID10 or ID02.
        :param series_id: ID of series to analyze in Xdata.meta dataframe.
        :param filename: output filename
        """
        to_h5(self, series_id, filename, **kwargs)
